#include "ehs/io/socket/EHC.h"
#include "ehs/io/socket/ehc/NetEnc.h"
#include "ehs/io/socket/ehc/NetSys.h"
#include "ehs/io/socket/ehc/NetEnd.h"
#include "ehs/Log.h"
#include "ehs/PRNG.h"


namespace ehs
{
	const Version EHC::version(1, 0, 0);
	const UInt_64 EHC::internalSys = Str_8::Hash_64("Internal");
	const UInt_64 EHC::connectOp = Str_8::Hash_64("Connect");
	const UInt_64 EHC::connectedOp = Str_8::Hash_64("Connected");
	const UInt_64 EHC::rejectedOp = Str_8::Hash_64("Rejected");
	const UInt_64 EHC::disconnectOp = Str_8::Hash_64("Disconnect");
	const UInt_64 EHC::disconnectedOp = Str_8::Hash_64("Disconnected");
	const UInt_64 EHC::statusUpdateOp = Str_8::Hash_64("StatusUpdate");
	const UInt_64 EHC::pingOp = Str_8::Hash_64("Ping");
	const UInt_64 EHC::pongOp = Str_8::Hash_64("Pong");
	const UInt_64 EHC::latencyOp = Str_8::Hash_64("Latency");
	const UInt_64 EHC::receivedOp = Str_8::Hash_64("Received");

	EHC::~EHC()
	{
		Release();
	}

	EHC::EHC()
		: udp(AddrType::IPV6), disposition(EndDisp::ENDPOINT), hashName(0), dropPackets(false), buffer(nullptr), bufferSize(0),
		maxEndpoints(0), lastTSC(0), delta(0.0f), maxTimeout(5.0f), resendRate(0.5f), connectCb(nullptr),
		connectedCb(nullptr), activeCb(nullptr), disconnectCb(nullptr), disconnectedCb(nullptr), timeoutCb(nullptr)
	{
	}

	EHC::EHC(const Version &ver, Str_8 name, const UInt_64 maxEndpoints)
		: udp(AddrType::IPV6), appVer(ver), disposition(EndDisp::ENDPOINT), hashName(name.Hash_64()), name((Str_8&&)name),
		dropPackets(false), buffer(nullptr), bufferSize(0), maxEndpoints(maxEndpoints), lastTSC(CPU::GetTSC()),
		delta(0.0f), maxTimeout(5.0f), resendRate(0.5f), connectCb(nullptr), connectedCb(nullptr), activeCb(nullptr),
		disconnectCb(nullptr), disconnectedCb(nullptr), timeoutCb(nullptr)
	{
		Initialize();
	}

	EHC::EHC(const Version &ver, Str_8 name)
		: udp(AddrType::IPV6), appVer(ver), disposition(EndDisp::ENDPOINT), hashName(name.Hash_64()), name((Str_8&&)name),
		dropPackets(false), buffer(nullptr), bufferSize(0), maxEndpoints(0), lastTSC(CPU::GetTSC()),
		delta(0.0f), maxTimeout(5.0f), resendRate(0.5f), connectCb(nullptr), connectedCb(nullptr), activeCb(nullptr),
		disconnectCb(nullptr), disconnectedCb(nullptr), timeoutCb(nullptr)
	{
		Initialize();
	}

	EHC::EHC(EHC &&sock) noexcept
		: udp((UDP&&)sock.udp), appVer(sock.appVer), disposition(sock.disposition), hashName(sock.hashName),
		  name((Str_8&&)sock.name), dropPackets(sock.dropPackets), buffer(sock.buffer), bufferSize(sock.bufferSize),
		  encryptions((Array<NetEnc *> &&)sock.encryptions), systems((Array<NetSys *> &&)sock.systems),
		  maxEndpoints(sock.maxEndpoints), lastTSC(sock.lastTSC), delta(sock.delta), maxTimeout(sock.maxTimeout),
		  resendRate(sock.resendRate), connectCb(sock.connectCb), connectedCb(sock.connectedCb), activeCb(sock.activeCb),
		  disconnectCb(sock.disconnectCb), disconnectedCb(sock.disconnectedCb), timeoutCb(sock.timeoutCb),
		  endpoints((Vector<NetEnd *> &&)sock.endpoints)
	{
		for (UInt_64 i = 0; i < encryptions.Size(); ++i)
			endpoints[i]->owner	= this;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			endpoints[i]->owner = this;

		sock.appVer = {};
		sock.disposition = EndDisp::ENDPOINT;
		sock.hashName = 0;
		sock.dropPackets = false;
		sock.buffer = nullptr;
		sock.bufferSize = 0;
		sock.maxEndpoints = 0;
		sock.lastTSC = 0;
		sock.delta = 0.0f;
		sock.maxTimeout = 5.0f;
		sock.resendRate = 0.5f;
		sock.connectCb = nullptr;
		sock.connectedCb = nullptr;
		sock.activeCb = nullptr;
		sock.disconnectCb = nullptr;
		sock.disconnectedCb = nullptr;
		sock.timeoutCb = nullptr;
	}

	EHC::EHC(const EHC &sock)
		: udp(sock.udp), appVer(sock.appVer), disposition(sock.disposition), hashName(sock.hashName), name(sock.name),
		dropPackets(sock.dropPackets), buffer(nullptr), bufferSize(0), maxEndpoints(sock.maxEndpoints),
		lastTSC(CPU::GetTSC()), delta(0.0f), maxTimeout(sock.maxTimeout), resendRate(sock.resendRate),
		connectCb(nullptr), connectedCb(nullptr), activeCb(nullptr), disconnectCb(nullptr), disconnectedCb(nullptr),
		timeoutCb(nullptr)
	{
		Initialize();
	}

	EHC &EHC::operator=(EHC &&sock) noexcept
	{
		if (this == &sock)
			return *this;

		udp = (UDP&&)sock.udp;
		appVer = sock.appVer;
		disposition = sock.disposition;
		hashName = sock.hashName;
		name = (Str_8&&)sock.name;
		dropPackets = sock.dropPackets;

		delete[] buffer;
		buffer = sock.buffer;

		bufferSize = sock.bufferSize;

		encryptions = (Array<NetEnc *> &&)sock.encryptions;
		for (UInt_64 i = 0; i < encryptions.Size(); ++i)
			encryptions[i]->owner = this;

		systems = (Array<NetSys *> &&)sock.systems;
		maxEndpoints = sock.maxEndpoints;
		lastTSC = sock.lastTSC;
		delta = sock.delta;
		maxTimeout = sock.maxTimeout;
		resendRate = sock.resendRate;
		connectCb = sock.connectCb;
		connectedCb = sock.connectedCb;
		activeCb = sock.activeCb;
		disconnectCb = sock.disconnectCb;
		disconnectedCb = sock.disconnectedCb;
		timeoutCb = sock.timeoutCb;

		endpoints = (Vector<NetEnd*>&&)sock.endpoints;
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			endpoints[i]->owner = this;

		sock.appVer = {};
		sock.disposition = EndDisp::ENDPOINT;
		sock.hashName = 0;
		sock.dropPackets = false;
		sock.buffer = nullptr;
		sock.bufferSize = 0;
		sock.maxEndpoints = 0;
		sock.lastTSC = 0;
		sock.delta = 0.0f;
		sock.maxTimeout = 5.0f;
		sock.resendRate = 0.5f;
		sock.connectCb = nullptr;
		sock.connectedCb = nullptr;
		sock.activeCb = nullptr;
		sock.disconnectCb = nullptr;
		sock.disconnectedCb = nullptr;
		sock.timeoutCb = nullptr;

		return *this;
	}

	EHC &EHC::operator=(const EHC &sock)
	{
		if (this == &sock)
			return *this;

		udp = sock.udp;
		appVer = sock.appVer;
		disposition = sock.disposition;
		hashName = sock.hashName;
		name = sock.name;
		dropPackets = sock.dropPackets;

		delete[] buffer;
		buffer = nullptr;

		bufferSize = 0;
		systems = Array<NetSys *>();
		maxEndpoints = sock.maxEndpoints;
        lastTSC = 0;
        delta = 0.0f;
		maxTimeout = sock.maxTimeout;
        resendRate = sock.resendRate;
		connectCb = nullptr;
		connectedCb = nullptr;
		activeCb = nullptr;
		disconnectCb = nullptr;
		disconnectedCb = nullptr;
		timeoutCb = nullptr;
		endpoints = Vector<NetEnd*>();

		Initialize();

		return *this;
	}

	void EHC::Initialize()
	{
		if (!udp.IsValid())
			return;

		udp.SetBlocking(false);

		if (udp.GetLocalAddressType() == AddrType::IPV4)
		{
			buffer = new Byte[EHS_IPV4_UDP_PAYLOAD];
			bufferSize = EHS_IPV4_UDP_PAYLOAD;
		}
		else if (udp.GetLocalAddressType() == AddrType::IPV6)
		{
			buffer = new Byte[EHS_IPV6_UDP_PAYLOAD];
			bufferSize = EHS_IPV6_UDP_PAYLOAD;
		}
	}

	void EHC::Release()
	{
		if (!udp.IsValid())
			return;

		disposition = EndDisp::ENDPOINT;

		delete[] buffer;
        buffer = nullptr;
        bufferSize = 0;

		Serializer payload(Endianness::LE);
		payload.Write<UInt_64>(0);

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetStatus() != Status::PENDING)
				endpoints[i]->Send(false, true, false, internalSys, disconnectOp, payload);

			delete endpoints[i];
		}
		endpoints.Clear();

		for (UInt_64 i = 0; i < systems.Size(); ++i)
			delete systems[i];

		systems.Clear();

		udp.Release();
	}

	void EHC::Bind(AddrType newType, const Str_8& newAddress, const UInt_16 newPort)
	{
		disposition = EndDisp::SERVICE;

		udp.Bind(newType, newAddress, newPort);
	}

	void EHC::Connect(AddrType rType, const Str_8& rAddress, const UInt_16 rPort, Serializer<UInt_64> data)
	{
		if (!udp.IsValid())
			return;

		Serializer payload(Endianness::LE);
		payload.WriteStr(name);
		payload.Write(CPU::GetArchitecture());
		payload.WriteVersion(version);
		payload.WriteVersion(appVer);
		payload.WriteSer((Serializer<UInt_64>&&)data);

		NetEnd* end = new NetEnd(rType, rAddress, rPort);
		end->owner = this;
		end->Send(false, true, false, "Internal", "Connect", payload);

		endpoints.Push(end);
	}

	bool EHC::Disconnect(const EndDisp endDisp, const Char_8 token[64], const Str_8& msg)
	{
		if (!udp.IsValid())
			return false;

		NetEnd* end = GetEndpoint(endDisp, token);
		if (!end)
			return false;

		Serializer<> payload(Endianness::LE);
		payload.WriteStr(msg);

		end->Send(false, true, false, internalSys, disconnectOp, payload);

		return true;
	}

	void EHC::Broadcast(const EndDisp endDisp, const Status endStatus, const bool deltaLocked, const UInt_64 encHashId,
			const bool ensure, const UInt_64 sysHashId, const UInt_64 opHashId, const Serializer<UInt_64> &payload)
	{
		if (!udp.IsValid())
			return;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() == endStatus)
				endpoints[i]->Send(deltaLocked, encHashId, ensure, sysHashId, opHashId, payload);
		}
	}

	void EHC::Broadcast(const EndDisp endDisp, const Status endStatus, const bool deltaLocked, const Str_8 &encId,
						   const bool ensure, const Str_8 &sysId, const Str_8 &opId,
						   const Serializer<UInt_64> &payload)
	{
		Broadcast(endDisp, endStatus, deltaLocked, encId.Hash_64(), ensure, sysId.Hash_64(), opId.Hash_64(), payload);
	}

	void EHC::Poll()
	{
		if (!udp.IsValid())
			return;

        UInt_64 newTSC = CPU::GetTSC();
        delta = (float)(newTSC - lastTSC) / (float)CPU::GetTSC_Freq();
        lastTSC = newTSC;

		AddrType rType = AddrType::IPV6;
		Str_8 rAddress;
		UInt_16 rPort = 0;

		UInt_16 received;

		while ((received = udp.Receive(&rType, &rAddress, &rPort, buffer, bufferSize)))
		{
			Serializer<> payload(Endianness::LE, buffer, received);

			UInt_64 encHashId = payload.Read<UInt_64>();
			if (encHashId)
			{
				NetEnc* enc = GetEncryption(encHashId);
				if (!enc)
					continue;

				enc->Encrypt(&payload[payload.GetOffset()], payload.Size() - payload.GetOffset());
			}

            payload.SetOffset(0);

			Header header = payload.Read<Header>();

			if (!header.ensure && !header.token[0] && header.system == internalSys && header.op == connectOp)
			{
				Str_8 endId = payload.ReadStr();
				Architecture rArch = payload.Read<Architecture>();

				NetEnd* end = new NetEnd(header.disposition, (Str_8&&)endId, rArch, rType, rAddress, rPort);
				end->owner = this;
				end->SetStatus(Status::PENDING);

				Serializer sPayload(Endianness::LE);

				Version rVer = payload.ReadVersion();
				if (rVer != version)
				{
					sPayload.WriteStr<Char_8, UInt_64>("Your Event Horizon Socket Layer version " +
						Str_8::FromNum(rVer.major) + "." + Str_8::FromNum(rVer.minor) + "." + Str_8::FromNum(rVer.patch) +
						" does not match remote endpoint version " +
						Str_8::FromNum(version.major) + "." + Str_8::FromNum(version.minor) + "." + Str_8::FromNum(version.patch) +
						". Connection rejected.");

					end->Send(false, true, false, internalSys, rejectedOp, sPayload);
					continue;
				}

				Version rAppVer = payload.ReadVersion();
				if (rAppVer != appVer)
				{
					sPayload.WriteStr<Char_8, UInt_64>("Your application version " +
						Str_8::FromNum(rAppVer.major) + "." + Str_8::FromNum(rAppVer.minor) + "." + Str_8::FromNum(rAppVer.patch) +
						" does not match remote endpoint application version " +
						Str_8::FromNum(appVer.major) + "." + Str_8::FromNum(appVer.minor) + "." + Str_8::FromNum(appVer.patch) +
						". Connection rejected.");

					end->Send(false, true, false, internalSys, rejectedOp, sPayload);
					continue;
				}

				GenerateToken(end->token);

				if (connectCb && !connectCb(this, &end, {Endianness::LE, &payload[payload.GetOffset()], payload.Size() - payload.GetOffset()}))
				{
					sPayload.WriteStr<Char_8, UInt_64>("Connection rejected.");
					end->Send(false, true, false, internalSys, rejectedOp, sPayload);
					continue;
				}

				endpoints.Push(end);

				sPayload.Write(CPU::GetArchitecture());
				sPayload.WriteStr(name);

				UInt_64 active = GetEndpointsCount(EndDisp::ENDPOINT, Status::ACTIVE);

				if (maxEndpoints && active >= maxEndpoints)
				{
					end->SetStatus(Status::IN_LOCAL_QUEUE);

					UpdateQueue(active);

					sPayload.Write(Status::IN_REMOTE_QUEUE);
					sPayload.Write(end->GetQueueSlot());
				}
				else
				{
					end->SetStatus(Status::ACTIVE);

					if (activeCb)
						activeCb(this, end);

					sPayload.Write(Status::ACTIVE);
					sPayload.Write(0);
				}

				end->Send(false, true, false, internalSys, connectedOp, sPayload);
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == connectedOp)
			{
				NetEnd* end = GetEndpoint(rAddress, rPort);
				if (!end || end->GetStatus() != Status::PENDING)
					continue;

				Str_8 endId = payload.ReadStr();
				Architecture arch = payload.Read<Architecture>();

				*end = NetEnd(header.disposition, (Str_8&&)endId, arch, rType, rAddress, rPort);
				end->owner = this;

				Util::Copy(end->token, header.token, 64);

				end->SetStatus(payload.Read<Status>());
				end->SetQueueSlot(payload.Read<UInt_64>());

				if (connectedCb)
					connectedCb(this, end);
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == rejectedOp)
			{
				if (!RemoveEndpoint(rAddress, rPort))
					continue;

				Str_8 msg = payload.ReadStr<Char_8, UInt_64>();
				if (msg.Size())
					EHS_LOG_INT(LogType::INFO, 3, msg);
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == disconnectOp)
			{
				NetEnd* end = GetEndpoint(header.disposition, header.token);
				if (!end)
					continue;

				end->Send(false, true, false, internalSys, disconnectedOp, {});

				if (disconnectCb)
					disconnectCb(this, end, {Endianness::LE, &payload[payload.GetOffset()], payload.Size() - payload.GetOffset()});

				RemoveEndpoint(header.disposition, end->token);

				UpdateQueue();
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == disconnectedOp)
			{
				NetEnd* end = GetEndpoint(header.disposition, header.token);
				if (!end)
					continue;

				if (disconnectedCb)
					disconnectedCb(this, end);

				RemoveEndpoint(end);
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == statusUpdateOp)
			{
				NetEnd* end = GetEndpoint(header.disposition, header.token);
				if (!end)
					continue;

				Status newStatus = payload.Read<Status>();
				UInt_64 newSlot = payload.Read<UInt_64>();

				if (end->GetStatus() == Status::ACTIVE)
				{
					if (activeCb)
						activeCb(this, end);

					//EHS_LOG_INT(LogType::INFO, 5, "Your connection status to " + end->GetId() + " has now become active.");
				}
				else if (end->GetStatus() == Status::IN_REMOTE_QUEUE && newStatus == Status::IN_REMOTE_QUEUE)
				{
					//EHS_LOG_INT(LogType::INFO, 5, "Your queue slot for " + end->GetId() + " is now " + newSlot + ".");
				}

				end->SetStatus(newStatus);
				end->SetQueueSlot(newSlot);
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == pingOp)
			{
				NetEnd* end = GetEndpoint(header.disposition, header.token);
				if (!end)
					continue;

				end->SetDeltaRate(payload.Read<float>());
                end->Pong(delta);
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == pongOp)
			{
				NetEnd* end = GetEndpoint(header.disposition, header.token);
				if (!end)
					continue;

				end->SetDeltaRate(payload.Read<float>());
                end->SendLatency();
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == latencyOp)
			{
				NetEnd* end = GetEndpoint(header.disposition, header.token);
				if (!end)
					continue;

				end->SetLatency(payload.Read<float>());
			}
			else if (!header.ensure && header.token[0] && header.system == internalSys && header.op == receivedOp)
			{
				NetEnd* end = GetEndpoint(header.disposition, header.token);
				if (!end)
					continue;

                const UInt_64 msgId = payload.Read<UInt_64>();
                const UInt_64 fragment = payload.Read<UInt_64>();

				end->RemoveInsurance(msgId, fragment);
			}
			else if (header.token[0])
			{
				NetEnd* end = GetEndpoint(header.disposition, header.token);
				if (!end)
					continue;

				if (dropPackets && !header.ensure && header.id < end->GetNextRecvId())
                {
                    EHS_LOG_INT(LogType::INFO, 6, "Old packet intentionally dropped.");
                    continue;
                }

                if (header.ensure)
                {
                    Serializer sPayload(Endianness::LE);
                    sPayload.Write(header.id);
                    sPayload.Write(header.fragment);

                    end->Send(false, true, false, internalSys, receivedOp, sPayload);
                }

				end->AddReceived(
					header,
					Serializer<>(Endianness::LE, &payload[payload.GetOffset()], payload.Size() - payload.GetOffset())
				);
			}
			else
			{
				EHS_LOG_INT(LogType::INFO, 7, "Corrupted packet.");
			}
		}

		PollEndpoints();
	}

	bool EHC::IsInitialized() const
	{
		return udp.IsValid();
	}

	AddrType EHC::GetLocalAddressType() const
	{
		return udp.GetLocalAddressType();
	}

	Str_8 EHC::GetLocalAddress() const
	{
		return udp.GetLocalAddress();
	}

	UInt_16 EHC::GetLocalPort() const
	{
		return udp.GetLocalPort();
	}

	bool EHC::IsBound() const
	{
		return udp.IsBound();
	}

	Version EHC::GetVersion()
	{
		return version;
	}

	Version EHC::GetAppVersion() const
	{
		return appVer;
	}

	EndDisp EHC::GetDisposition() const
	{
		return disposition;
	}

	UInt_64 EHC::GetHashName() const
	{
		return hashName;
	}

	Str_8 EHC::GetName() const
	{
		return name;
	}

	void EHC::EnableDropPackets(const bool enable)
	{
		dropPackets = enable;
	}

	bool EHC::IsDropPacketsEnabled() const
	{
		return dropPackets;
	}

	bool EHC::HasEncryption(UInt_64 encHashId) const
	{
		for (UInt_64 i = 0; i < encryptions.Size(); ++i)
			if (encryptions[i]->GetHashId() == encHashId)
				return true;

		return false;
	}

	bool EHC::HasEncryption(const Str_8& encId) const
	{
		return HasEncryption(encId.Hash_64());
	}

	bool EHC::AddEncryption(NetEnc* enc)
	{
		if (HasEncryption(enc->GetHashId()))
			return false;

		enc->owner = this;
		encryptions.Push(enc);

		return true;
	}

	NetEnc* EHC::GetEncryption(UInt_64 encHashId) const
	{
		for (UInt_64 i = 0; i < encryptions.Size(); ++i)
			if (encryptions[i]->GetHashId() == encHashId)
				return encryptions[i];

		return nullptr;
	}

	NetEnc* EHC::GetEncryption(const Str_8& encId) const
	{
		return GetEncryption(encId.Hash_64());
	}

	bool EHC::HasSystem(const UInt_64 sysHashId) const
	{
		if (internalSys == sysHashId)
			return true;

		for (UInt_64 i = 0; i < systems.Size(); ++i)
			if (systems[i]->GetHashId() == sysHashId)
				return true;

		return false;
	}

	bool EHC::HasSystem(const Str_8& sysId) const
	{
		return HasSystem(sysId.Hash_64());
	}

	bool EHC::AddSystem(NetSys *sys)
	{
		if (HasSystem(sys->GetHashId()))
			return false;

		systems.Push(sys);

		return true;
	}

	NetSys* EHC::GetSystem(const UInt_64 sysHashId) const
	{
		for (UInt_64 i = 0; i < systems.Size(); ++i)
			if (systems[i]->GetHashId() == sysHashId)
				return systems[i];

		return nullptr;
	}

	NetSys* EHC::GetSystem(const Str_8& sysId) const
	{
		return GetSystem(sysId.Hash_64());
	}

	bool EHC::HasEndpoint(const EndDisp endDisp, const Status endStatus, const Char_8 token[64]) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() != endStatus)
				continue;

			if (Util::Compare(endpoints[i]->token, token, 64))
				return true;
		}

		return false;
	}

	bool EHC::HasEndpoint(const EndDisp endDisp, const Status endStatus, const UInt_64 hashName) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() != endStatus)
				continue;

			if (endpoints[i]->GetHashName() == hashName)
				return true;
		}

		return false;
	}

	bool EHC::HasEndpoint(const EndDisp endDisp, const Status endStatus, const Str_8 &id) const
	{
		return HasEndpoint(endDisp, endStatus, id.Hash_64());
	}

	bool EHC::HasEndpoint(const EndDisp endDisp, const Char_8 token[64]) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (Util::Compare(endpoints[i]->token, token, 64))
				return true;
		}

		return false;
	}

	bool EHC::HasEndpoint(const EndDisp endDisp, const UInt_64 hashName) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetHashName() == hashName)
				return true;
		}

		return false;
	}

	bool EHC::HasEndpoint(const EndDisp endDisp, const Str_8 &id) const
	{
		return HasEndpoint(endDisp, id.Hash_64());
	}

	bool EHC::HasEndpoint(const Char_8 token[64]) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (Util::Compare(endpoints[i]->token, token, 64))
				return true;

		return false;
	}

	bool EHC::HasEndpoint(UInt_64 hashName) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetHashName() == hashName)
				return true;

		return false;
	}

	bool EHC::HasEndpoint(const Str_8 &id) const
	{
		return HasEndpoint(id.Hash_64());
	}

	bool EHC::HasEndpoint(const Str_8& rAddress, const UInt_16 rPort) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetAddress() == rAddress && endpoints[i]->GetPort() == rPort)
				return true;

		return false;
	}

	NetEnd* EHC::GetEndpoint(const EndDisp endDisp, const Status endStatus, const Char_8 token[64]) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() != endStatus)
				continue;

			if (Util::Compare(endpoints[i]->token, token, 64))
				return endpoints[i];
		}

		return nullptr;
	}

	NetEnd *EHC::GetEndpoint(const EndDisp endDisp, const Status endStatus, const UInt_64 hashName) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() != endStatus)
				continue;

			if (endpoints[i]->GetHashName() == hashName)
				return endpoints[i];
		}

		return nullptr;
	}

	NetEnd *EHC::GetEndpoint(const EndDisp endDisp, const Status endStatus, const Str_8 &id) const
	{
		return GetEndpoint(endDisp, endStatus, id.Hash_64());
	}

	NetEnd *EHC::GetEndpoint(const EndDisp endDisp, const Char_8 token[64]) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (Util::Compare(endpoints[i]->token, token, 64))
				return endpoints[i];
		}

		return nullptr;
	}

	NetEnd *EHC::GetEndpoint(const EndDisp endDisp, const UInt_64 hashName) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetHashName() == hashName)
				return endpoints[i];
		}

		return nullptr;
	}

	NetEnd *EHC::GetEndpoint(const EndDisp endDisp, const Str_8 &id) const
	{
		return GetEndpoint(endDisp, id.Hash_64());
	}

	NetEnd *EHC::GetEndpoint(const Str_8& rAddress, const UInt_16 rPort) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetAddress() == rAddress && endpoints[i]->GetPort() == rPort)
				return endpoints[i];

		return nullptr;
	}

	Array<NetEnd *> EHC::GetEndpoints(const EndDisp endDisp, const Status endStatus)
	{
		Array<NetEnd*> result(endpoints.Size());
		UInt_64 count = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() == endStatus)
				result[count++] = endpoints[i];
		}

		result.Resize(count);

		return result;
	}

	Array<NetEnd *> EHC::GetEndpoints(const EndDisp endDisp)
	{
		Array<NetEnd*> result(endpoints.Size());
		UInt_64 count = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetDisposition() == endDisp)
				result[count++] = endpoints[i];

		result.Resize(count);

		return result;
	}

	UInt_64 EHC::GetEndpointsCount(const EndDisp endDisp, const Status endStatus)
	{
		UInt_64 count = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() == endStatus)
				++count;
		}

		return count;
	}

	UInt_64 EHC::GetEndpointsCount(const EndDisp endDisp)
	{
		UInt_64 count = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetDisposition() == endDisp)
				++count;

		return count;
	}

	UInt_64 EHC::GetMaxEndpoints() const
	{
		return maxEndpoints;
	}

	void EHC::SetMaxTimeout(const float seconds)
	{
		maxTimeout = seconds;
	}

	float EHC::GetMaxTimeout() const
	{
		return maxTimeout;
	}

    void EHC::SetResendRate(const float seconds)
    {
        resendRate = seconds;
    }

    float EHC::GetResendRate() const
    {
        return resendRate;
    }

	void EHC::SetConnectCb(const ConnectCb cb)
	{
		connectCb = cb;
	}

	void EHC::SetConnectedCb(const ConnectedCb cb)
	{
		connectedCb = cb;
	}

	void EHC::SetActiveCb(const ActiveCb cb)
	{
		activeCb = cb;
	}

	void EHC::SetDisconnectCb(const DisconnectCb cb)
	{
		disconnectCb = cb;
	}

	void EHC::SetDisconnectedCb(const DisconnectedCb cb)
	{
		disconnectedCb = cb;
	}

	void EHC::SetTimeoutCb(const TimeoutCb cb)
	{
		timeoutCb = cb;
	}

	void EHC::GenerateToken(Char_8 in[64])
	{
		PRNG_u64 rng(CPU::GetTSC());

		for (UInt_64 i = 0; i < 8; ++i)
		{
			do
				((UInt_64*)in)[i] = rng.Generate();
			while (!i && ((UInt_64*)in)[i] == 0);
		}

		if (HasEndpoint(in))
			GenerateToken(in);
	}

	void EHC::UpdateQueue(UInt_64 active)
	{
		UInt_64 slot = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetStatus() == Status::IN_LOCAL_QUEUE)
			{
				if (active < maxEndpoints)
				{
					endpoints[i]->SetStatus(Status::ACTIVE);
					endpoints[i]->SetQueueSlot(0);

					Serializer payload(Endianness::LE);
					payload.Write(Status::ACTIVE);
					payload.Write(0);

					endpoints[i]->Send(false, true, false, internalSys, statusUpdateOp, payload);

					if (activeCb)
						activeCb(this, endpoints[i]);

					++active;
				}
				else
				{
					if (endpoints[i]->GetQueueSlot() != slot)
					{
						Serializer payload(Endianness::LE);
						payload.Write(Status::IN_REMOTE_QUEUE);
						payload.Write(slot);

						endpoints[i]->Send(false, true, false, internalSys, statusUpdateOp, payload);

						endpoints[i]->SetQueueSlot(slot++);
					}
					else
					{
						++slot;
					}
				}
			}
		}
	}

	void EHC::UpdateQueue()
	{
		UpdateQueue(GetEndpointsCount(EndDisp::ENDPOINT, Status::ACTIVE));
	}

	bool EHC::RemoveEndpoint(const EndDisp endDisp, const Char_8 token[64])
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (Util::Compare(endpoints[i]->token, token, 64))
			{
				delete endpoints[i];

				if (i != endpoints.Size() - 1)
					endpoints.Swap(i, endpoints.End());

				endpoints.Pop();

				return true;
			}
		}

		return false;
	}

	bool EHC::RemoveEndpoint(const Str_8& rAddress, const UInt_16 rPort)
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetAddress() == rAddress && endpoints[i]->GetPort() == rPort)
			{
				delete endpoints[i];

				if (i != endpoints.Size() - 1)
					endpoints.Swap(i, endpoints.End());

				endpoints.Pop();

				return true;
			}
		}

		return false;
	}

	bool EHC::RemoveEndpoint(const NetEnd* const end)
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i] == end)
			{
				delete endpoints[i];

				if (i != endpoints.Size() - 1)
					endpoints.Swap(i, endpoints.End());

				endpoints.Pop();

				return true;
			}
		}

		return false;
	}

	void EHC::PollEndpoints()
	{
		UInt_64 i = 0;
		while (i < endpoints.Size())
		{
			endpoints[i]->Poll(delta);

			if (endpoints[i]->GetStatus() == Status::PENDING)
			{
				if (endpoints[i]->GetTimeout() >= maxTimeout)
				{
					if (timeoutCb)
						timeoutCb(this, endpoints[i]);

					delete endpoints[i];

					if (i != endpoints.Size() - 1)
						endpoints.Swap(i, endpoints.End());

					endpoints.Pop();

					continue;
				}
			}
			else
			{
				if (endpoints[i]->GetTimeout() >= maxTimeout)
				{
					if (timeoutCb)
						timeoutCb(this, endpoints[i]);

					delete endpoints[i];

					if (i != endpoints.Size() - 1)
						endpoints.Swap(i, endpoints.End());

					endpoints.Pop();

					UpdateQueue();

					continue;
				}

				Vector<NetFrags>* frags = endpoints[i]->GetReceived();

                UInt_64 f = 0;
                while (f < frags->Size())
                {
                    if (!(*frags)[f].IsComplete())
                    {
                        ++f;
                        continue;
                    }

                    Packet packet = (*frags)[f].Combine();

					NetSys* sys = GetSystem(packet.header.system);
                    if (!sys)
                    {
                        ++f;
                        continue;
                    }

                    sys->Execute(this, endpoints[i], packet.header.op, packet.payload);

                    frags->Swap(f, frags->End());
                    frags->Pop();
                }
			}

			++i;
		}
	}
}
